# Copyright © 2023 Daniele Tricoli <eriol@mornie.org>
# SPDX-License-Identifier: BSD-3-Clause

"""A simple tool to send messages into FreakWAN over Bluetooth low energy."""

import asyncio
import sys

import bleak

from .cli import get_cli


def run() -> None:
    """Main entrypoint."""
    try:
        asyncio.run(get_cli())
    except (RuntimeError, bleak.exc.BleakError) as e:
        sys.exit(str(e))
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    run()



# Copyright © 2023 Daniele Tricoli <eriol@mornie.org>
# SPDX-License-Identifier: BSD-3-Clause

"""CLI related stuff for freakble."""

import logging
import sys

import asyncclick as click

from . import __version__, ble


def ble_receive_callback(data: bytes):
    """Print data received from Bluetooth LE."""
    click.echo(data)


@click.group()
@click.option(
    "--adapter", default="hci0", show_default="hci0", type=str, help="ble adapter"
)
@click.pass_context
def cli(ctx, adapter):
    """A simple tool to send messages into FreakWAN."""
    ctx.ensure_object(dict)
    ctx.obj["ADAPTER"] = adapter


@cli.command()
@click.option("--loop", is_flag=True, default=False, help="send forever the message")
@click.option(
    "--device",
    required=True,
    type=str,
    envvar="FREAKBLE_DEVICE",
    help="ble device address",
)
@click.option(
    "--sleep-time",
    default=1,
    show_default="1 sec",
    type=float,
    help="sleep between messages sent with --loop",
)
@click.option(
    "--timeout",
    default=10,
    show_default="10 secs",
    type=float,
    help="Bluetooth LE connection timeout",
)
@click.argument("words", type=str, nargs=-1)
@click.pass_context
async def send(ctx, words, device, loop, sleep_time, timeout):
    """Send one or more words over BLE to a specific device."""
    msg = " ".join(words)
    logging.info(f"Connecting to {device}...")
    try:
        await ble.send_text(
            ctx.obj["ADAPTER"],
            msg,
            device,
            loop,
            sleep_time,
            timeout,
            ble_receive_callback,
        )
    except AssertionError as e:
        click.echo(click.style(e, fg="red"))


@cli.command()
@click.option(
    "--scan-time", default=5, show_default="5 secs", type=float, help="scan duration"
)
@click.pass_context
async def scan(ctx, scan_time):
    """Scan to find BLE devices."""
    devices = await ble.scan(ctx.obj["ADAPTER"], scan_time)
    for device in devices:
        print(f"{device.address} (rssi:{device.rssi}) {device.name}")


@cli.command()
@click.option(
    "--device",
    required=True,
    type=str,
    envvar="FREAKBLE_DEVICE",
    help="ble device address",
)
@click.option(
    "--timeout",
    default=10,
    show_default="10 secs",
    type=float,
    help="Bluetooth LE connection timeout",
)
@click.pass_context
async def repl(ctx, device, timeout):
    """Start a REPL with the device."""
    click.echo(f"freakble {__version__} on {sys.platform}")
    click.echo(f"Connecting to {device}...")
    try:
        await ble.repl_loop(ctx.obj["ADAPTER"], device, timeout)
    except AssertionError as e:
        click.echo(click.style(e, fg="red"))


@cli.command()
@click.option(
    "--device",
    default=None,
    type=str,
    envvar="FREAKBLE_DEVICE",
    help="ble device address",
)
@click.option(
    "--ble-connection-timeout",
    default=10,
    show_default="10 secs",
    type=float,
    help="Bluetooth LE connection timeout",
)
@click.pass_context
async def gui(ctx, device, ble_connection_timeout):
    """Start freakble GUI."""

    # Imported here so that the others commands will works if tkinter is not
    # installed. It's OK to blow up to let user know for now.
    from .gui import App

    with App() as app:
        app.config(ctx.obj["ADAPTER"], device, ble_connection_timeout)
        await app.run()


@cli.command()
async def version():
    """Return freakble version."""
    click.echo(f"freakble {__version__}")


def get_cli():
    """Return the CLI entrypoint."""
    return cli(auto_envvar_prefix="FREAKBLE", _anyio_backend="asyncio")



# Copyright © 2023 Daniele Tricoli <eriol@mornie.org>
# SPDX-License-Identifier: BSD-3-Clause

"""This module contains a simple GUI for freakble."""

import asyncio
import tkinter as tk
from datetime import datetime
from tkinter import ttk

try:
    from ttkthemes import ThemedTk

    ARE_THEMES_AVAILABLE = True
except ImportError:
    ARE_THEMES_AVAILABLE = False

from .ble import Client, scan

WINDOW_SIZE = "800x600"


class App:
    def config(self, adapter, device, ble_connection_timeout):
        self.adapter = adapter
        self.device = device
        self.ble_connection_timeout = ble_connection_timeout
        self.loop = asyncio.get_event_loop()

    async def run(self):
        if ARE_THEMES_AVAILABLE:
            self.window = MainWindow(self, theme="breeze")
        else:
            self.window = MainWindow(self)
        await self.window.show()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        self.window.destroy()


if ARE_THEMES_AVAILABLE:
    BaseWindow = type("BaseWindow", (ThemedTk,), {})
else:
    BaseWindow = type("BaseWindow", (tk.Tk,), {})


class MainWindow(BaseWindow):
    def __init__(self, app, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.title("freakble")
        self.geometry(WINDOW_SIZE)
        self.option_add("*Font", "12")

        self.app = app
        self.is_app_closing = False

        self.protocol("WM_DELETE_WINDOW", self.quit)

        self.make_ui()

        self.windows = {}

        for window in (ScanWindow, DeviceWindow):
            w = window(self.container, self)
            self.windows[window] = w
            w.grid(row=0, column=0, sticky="news")

        if self.app.device is None:
            self.show_window(ScanWindow)
        else:
            self.windows[DeviceWindow].connect()
            self.show_window(DeviceWindow)

    def make_ui(self):
        self.container = ttk.Frame(self)
        self.container.pack(side=tk.TOP, fill=tk.BOTH, expand=True)
        self.container.rowconfigure(0, weight=1)
        self.container.columnconfigure(0, weight=1)

    def quit(self):
        self.is_app_closing = True

    async def show(self):
        while not self.is_app_closing:
            self.update()
            await asyncio.sleep(0.1)

    def show_window(self, window):
        for w in self.windows:
            w.grid_forget(self)

        frame = self.windows[window]
        frame.tkraise()


class ScanWindow(ttk.Frame):
    def __init__(self, parent, main_window):
        super().__init__(parent)

        self.parent = parent
        self.main_window = main_window

        self.make_ui()

    def make_ui(self):
        # Make the frame expand to the whole window.
        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)

        self.frame_devices = ttk.LabelFrame(self, text="Devices", width=100, height=100)
        self.frame_devices.rowconfigure(0, weight=1)
        self.frame_devices.columnconfigure(0, weight=1)
        self.frame_devices.grid(row=0, column=0, sticky="news", padx=10, pady=10)

        self.listbox = tk.Listbox(
            self.frame_devices, width=40, height=10, selectmode=tk.SINGLE, bg="white"
        )
        self.listbox.bind("<<ListboxSelect>>", self.on_listbox_clicked)
        self.listbox.grid(row=0, column=0, sticky="news", padx=10, pady=10)

        self.frame_buttons = ttk.Frame(self, width=100)
        self.frame_buttons.grid(row=1, column=0, sticky="news")
        self.frame_buttons.rowconfigure(0, weight=1)
        self.frame_buttons.columnconfigure(0, weight=1)
        self.frame_buttons.columnconfigure(1, weight=1)
        self.button_scan = ttk.Button(
            self.frame_buttons,
            text="Scan",
            command=lambda: self.main_window.app.loop.create_task(
                self.on_scan_clicked()
            ),
        )
        self.button_scan.grid(row=0, column=0, sticky="news", padx=10, pady=10)

        self.button_connect = ttk.Button(
            self.frame_buttons,
            text="Connect",
            state=tk.DISABLED,
            command=self.on_button_connect_clicked,
        )
        self.button_connect.grid(row=0, column=1, sticky="news", padx=10, pady=10)

    async def on_scan_clicked(self):
        # Clear listbox.
        self.listbox.delete(0, self.listbox.size())

        self.button_connect["state"] = tk.DISABLED
        self.button_scan["state"] = tk.DISABLED
        self.button_scan.configure(text="Scanning...")

        devices = await scan(
            self.main_window.app.adapter,
            self.main_window.app.ble_connection_timeout,
        )
        for i, device in enumerate(devices):
            self.listbox.insert(
                i, f"{device.address} (rssi: {device.rssi}) {device.name}"
            )

        self.button_scan.configure(text="Scan")
        self.button_scan["state"] = tk.NORMAL

    def on_listbox_clicked(self, _):
        self.button_connect["state"] = tk.NORMAL

    def on_button_connect_clicked(self):
        device = self.listbox.get(self.listbox.curselection())
        self.main_window.app.device = device.split()[0]
        self.main_window.show_window(DeviceWindow)
        self.main_window.windows[DeviceWindow].connect()


class DeviceWindow(ttk.Frame):
    def __init__(self, parent, main_window):
        super().__init__(parent)

        self.main_window = main_window

        self.make_ui()

    def connect(self):
        self.task = self.main_window.app.loop.create_task(self.ble_loop())

    def make_ui(self):
        # Make the frame expand to the whole window.
        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)

        self.frame_text = ttk.Frame(self, relief="ridge", width=100, height=100)
        self.frame_text.rowconfigure(0, weight=1)
        self.frame_text.columnconfigure(0, weight=1)
        self.v_scrollbar = ttk.Scrollbar(self.frame_text)
        self.v_scrollbar.pack(side=tk.RIGHT, fill=tk.Y)
        self.text = tk.Text(
            self.frame_text,
            bg="white",
            width=100,
            height=100,
            yscrollcommand=self.v_scrollbar.set,
            state=tk.DISABLED,
        )
        self.text.pack(side=tk.TOP, fill=tk.X)
        self.v_scrollbar.config(command=self.text.yview)

        self.frame_text.grid(row=0, column=0, sticky="news")

        self.frame_send = ttk.Frame(self, width=100)
        self.frame_send.grid(row=1, column=0, sticky="news")
        self.frame_send.rowconfigure(1, weight=1)
        self.frame_send.columnconfigure(0, weight=1)

        self.entry = ttk.Entry(self.frame_send, width=100)
        self.entry.grid(row=1, column=0, sticky="news")
        self.entry.rowconfigure(0, weight=1)
        self.entry.columnconfigure(0, weight=1)
        self.entry.focus_set()
        self.entry.bind(
            "<Return>", lambda event: asyncio.ensure_future(self.on_entry_return(event))
        )
        self.button = ttk.Button(self.frame_send, text="⮞")
        self.button.grid(row=1, column=1, sticky="nesw")
        self.button.rowconfigure(1, weight=1)
        self.button.columnconfigure(1, weight=1)
        self.button.bind(
            "<Button-1>",
            lambda event: asyncio.ensure_future(self.on_entry_return(event)),
        )

    async def ble_loop(self):
        self.client = Client(self.main_window.app.adapter, self.main_window.app.device)
        self.client.set_receive_callback(self.on_ble_data_received)
        try:
            await self.client.connect(self.main_window.app.ble_connection_timeout)
            await self.client.start()
            await self.client.wait_until_disconnect()
        finally:
            await self.client.disconnect()

    async def send_over_ble(self, data):
        await self.client.send(bytes(data, "utf-8"))

    def on_ble_data_received(self, data):
        self.insert_text(f"{data}\n")

    def insert_text(self, text):
        now = datetime.now().strftime("%y/%m/%d %H:%M:%S")
        self.text["state"] = tk.NORMAL
        self.text.insert(tk.END, f"[{now}] {text}")
        self.text["state"] = tk.DISABLED
        # Scroll to the end.
        self.text.see(tk.END)

    async def on_entry_return(self, _):
        text = self.entry.get()
        await self.send_over_ble(text)
        self.insert_text(f"{text}\n")
        self.entry.delete(0, len(text))



# Copyright © 2023 Daniele Tricoli <eriol@mornie.org>
# SPDX-License-Identifier: BSD-3-Clause

"""This module contains a simple interactive shell for freakble."""

import asyncio
from itertools import cycle

from prompt_toolkit.completion import NestedCompleter
from prompt_toolkit.patch_stdout import patch_stdout
from prompt_toolkit.shortcuts import PromptSession

PROMPT_MESSAGE = "Φ] "

# These bandwidth values are for LYLIGO TTGO LoRa (TM) v2 1.6 and similar, the
# boards we use for FreakWAN.
BW_VALUES = [
    7800,
    10400,
    15600,
    20800,
    31250,
    41700,
    62500,
    62500,
    125000,
    250000,
    500000,
]

command_completer = NestedCompleter.from_nested_dict(
    {
        "!automsg": {"on": None, "off": None},
        "!bat": None,
        "!bw": dict(zip(map(str, BW_VALUES), cycle([None]))),
        "!cr": dict(zip(map(str, range(5, 9)), cycle([None]))),
        "!help": None,
        "!sp": dict(zip(map(str, range(6, 13)), cycle([None]))),
    }
)


class REPL:
    """Simple interactive shell able to send data over Bluetooth LE."""

    def __init__(self, client) -> None:
        # We use the client to set the callback and to send and receive data.
        # The ownership remains outside of this class.
        self._client = client
        self._client.set_receive_callback(self._on_ble_data_received)

    async def shell(self):
        """Display a prompt and handle user interaction."""
        session = PromptSession(PROMPT_MESSAGE, completer=command_completer)
        while True:
            try:
                with patch_stdout():
                    text = await session.prompt_async()
                    await self._client.send(bytes(text, "utf-8"))
            except (EOFError, KeyboardInterrupt):
                raise asyncio.CancelledError

    def _on_ble_data_received(self, data):
        """Print data received from Bluetooth LE."""
        print(f"{data}")



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['freakble']

package_data = \
{'': ['*']}

install_requires = \
['anyio>=3.6.2,<4.0.0',
 'asyncclick>=8.1.3.4,<9.0.0.0',
 'bleak>=0.19.5,<0.20.0',
 'prompt-toolkit>=3.0.36,<4.0.0']

extras_require = \
{'themes': ['ttkthemes>=3.2.2,<4.0.0']}

entry_points = \
{'console_scripts': ['freakble = freakble.__main__:run']}

setup_kwargs = {
    'name': 'freakble',
    'version': '0.6.0',
    'description': 'A simple tool to send messages into FreakWAN over Bluetooth low energy.',
    'long_description': '# freakble\n\nA simple tool to send messages and commands into [FreakWAN](https://github.com/antirez/freakwan)\nover Bluetooth low energy.\nIt\'s tested under Linux and Windows but it should work also on MacOS.\n\n## Installation\n\n### Using pipx\n\nThe best way to install freakble is using [pipx](https://pypa.github.io/pipx/):\n```console\n$ pipx install freakble\n```\n\nFor the GUI you can optionally install themes using:\n```console\n$ pipx install \'freakble[themes]\'\n```\n\n### Using pip\n\n*When using pip it\'s suggested to work inside a virtualenv.*\n\n```console\n$ python -m pip install freakble\n```\n\n### From source\n\nfreakble uses [Poetry](https://python-poetry.org) as dependency management and\npackaging tool, you need to install it first.\n\nThen:\n\n1. Clone this repository.\n2. From the root of the repository run:\n   ```console\n   $ poetry build\n   ```\n3. Install using pipx or pip (it\'s better to use pipx):\n   ```console\n   $ pipx install dist/freakble-0.1.0-py3-none-any.whl\n   ```\n\n## Usage\n\n```console\nUsage: freakble [OPTIONS] COMMAND [ARGS]...\n\n  A simple tool to send messages into FreakWAN.\n\nOptions:\n  --adapter TEXT  ble adapter  [default: (hci0)]\n  --help          Show this message and exit.\n\nCommands:\n  gui      Start freakble GUI.\n  repl     Start a REPL with the device.\n  scan     Scan to find BLE devices.\n  send     Send one or more words over BLE to a specific device.\n  version  Return freakble version.\n```\n\n### send\n\nThe `send` command is used to send a message to the board. You need to know the\naddress of the device.\nYou can specify the address of a device using the `--device` flag or the\nenvironment variable `FREAKBLE_DEVICE`.\n\nThe complete usage is:\n```console\nUsage: freakble send [OPTIONS] [WORDS]...\n\n  Send one or more words over BLE to a specific device.\n\nOptions:\n  --loop              send forever the message\n  --device TEXT       ble device address  [required]\n  --sleep-time FLOAT  sleep between messages sent with --loop  [default: (1\n                      sec)]\n  --timeout FLOAT     Bluetooth LE connection timeout  [default: (10 secs)]\n  --help              Show this message and exit.\n```\n\nFor example:\n\n```console\n$ freakble send --device AA:AA:AA:AA:AA:AA Hello, there!\n```\n\nwhere you have to substitute `AA:AA:AA:AA:AA:AA` with your device\'s address.\n\nThe `--loop` flag will make freakble to send continuosly the message until\n`CTRL + C` is pressed. The resend interval is defaults to 1 sec and can be\nchanged using `--sleep-time`.\n\n```console\n$ freakble send --device AA:AA:AA:AA:AA:AA --loop FREAKNET\n```\n\n![A photo of a LYLIGO TTGO LoRa v2 1.6 showing the text: you> FREAKNET in multiple lines.](extras/304f4bb6-4f51-4183-95b9-c329b9bf69ab.jpg)\n\nYou can use `FREAKBLE_DEVICE` environment variables to set the device address,\nand to not have to provide it in each commands that need a device address.\n\nFor example, using `send`, if one of your device is called `FW_vuzasu`\nyou can do:\n\n```console\n$ export FREAKBLE_DEVICE=$(freakble scan | grep FW_vuzasu | cut -d\' \' -f1)\n$ freakble send "La violenza è l\'ultimo rifugio degli incapaci. - Isaac Asimov"\n```\n\nFreakWAN supports commands starting with `!` character (see the list on\nFreakWAN README):\n\n```console\n$ freakble send \'!bat\'\nbattery 99%, 4.20 volts\n```\n\nIf you don\'t know what to say, you can always tempt the fates! :)\n```console\nfreakble send "$(fortune)"\n```\n\nPlease note that this command handle also disconnection from Bluetooth LE so it\nruns in a few seconds.\n\n### scan\n\nThe `scan` command is used to discover Bluetooth LE devices.\n\n```console\nUsage: freakble scan [OPTIONS]\n\n  Scan to find BLE devices.\n\nOptions:\n  --scan-time FLOAT  scan duration  [default: (5 secs)]\n  --help             Show this message and exit.\n```\n\nFor example:\n```\n$ freakble scan\nAB:AB:AB:AB:AB:AB (rssi:-52) FW_timatu\nAF:AF:AF:AF:AF:AF (rssi:-57) FW_vuzasu\n```\n\n### repl\n\nThe `repl` command connects to the specified device and stats an interactive\nshell with it.\nThe complete usage is:\n```\nUsage: freakble repl [OPTIONS]\n\n  Start a REPL with the device.\n\nOptions:\n  --device TEXT    ble device address  [required]\n  --timeout FLOAT  Bluetooth LE connection timeout  [default: (10 secs)]\n  --help           Show this message and exit.\n```\n\nFor example:\n\n```console\n$ export FREAKBLE_DEVICE=$(freakble scan | grep FW | cut -d\' \' -f1)\nfreakble 0.3.0a0 on linux\nConnecting to AB:AB:AB:AB:AB:AB...\nΦ]\n```\n\n`Φ]` is the freakble prompt.\n\nYou can then talk to the device remembering that commands start with `!` and\nthe text you write if it\'s not a command is sent as a message in the network.\n\nFor example, the following text is sent as a message in the network:\n```\nΦ] Hello there!\nΦ]\n```\n\nInstead commands make you able to get info or configure your FreakWAN node:\n```\nΦ] !help\nCommands: !automsg !pw !sp !cr !bw !freq !preset !ls !font !last !addkey !delkey !keys !usekey !nokey\nΦ] !bat\nbattery volts: 4.2\n```\n\nPressing `TAB` key or `!` will show the autocompletion menu.\n\nTo exit from the interactive shell use `CTRL + D` or `CTRL + C`\n\n### gui\n\nThe `gui` command starts a GUI of freakble. If you don\'t set the device address\nit will start with the following scan interface:\n\n![A screenshot of the scan inferface of the gui.](extras/gui_scan.png)\n\nafter choosing a device you can click to the connect button to get a graphical\nREPL:\n\n![A screenshot of the graphical REPL of freakble.](extras/gui_repl.png)\n\n\n## License\n\nfreakble is licensed under BSD-3-Clause license.\n',
    'author': 'Daniele Tricoli',
    'author_email': 'eriol@mornie.org',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/eriol/freakble',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# Copyright © 2023 Daniele Tricoli <eriol@mornie.org>
# SPDX-License-Identifier: BSD-3-Clause

"""BLE related stuff for freakble."""

import asyncio
import logging
from typing import Any, Callable

from bleak import BleakClient, BleakScanner

from .repl import REPL

__all__ = [
    "Client",
    "repl_loop",
    "scan",
    "send_text",
]

DEFAULT_TIMEOUT = 5.0
NORDIC_UART_SERVICE_UUID = "6e400001-b5a3-f393-e0a9-e50e24dcca9e"
NORDIC_UART_TX_CHARACTERISTIC = "6e400002-b5a3-f393-e0a9-e50e24dcca9e"
NORDIC_UART_RX_CHARACTERISTIC = "6e400003-b5a3-f393-e0a9-e50e24dcca9e"


async def scan(adapter: str, timeout: float = DEFAULT_TIMEOUT):
    """Scan for Bluetooth LE devices."""
    return await BleakScanner.discover(adapter=adapter, timeout=timeout)


class Client:
    """Simple client for UART devices."""

    def __init__(self, adapter: str, address: str) -> None:
        self.adapter = adapter
        self.address = address

        self._receive_callback = None
        self.disconnect_event = asyncio.Event()
        self._client = None

    async def connect(self, timeout: float = DEFAULT_TIMEOUT):
        device = await BleakScanner.find_device_by_address(
            self.address, adapter=self.adapter, timeout=timeout
        )
        if device is None:
            raise RuntimeError(f"device with address {self.address} not found")

        self._client = BleakClient(device, disconnected_callback=self.on_disconnect)
        await self._client.__aenter__()

    async def disconnect(self):
        if self._client is not None:
            await self.stop()
            await self._client.disconnect()

    async def start(self):
        if self._client is not None:
            await self._client.start_notify(NORDIC_UART_RX_CHARACTERISTIC, self.on_rx)

    async def stop(self):
        if self._client is not None:
            await self._client.stop_notify(NORDIC_UART_RX_CHARACTERISTIC)

    def set_receive_callback(self, callback: Callable[[Any], None]):
        self._receive_callback = callback

    def on_rx(self, characteristic, data):
        logging.debug(characteristic, data)
        data = data.decode("utf-8").rstrip()

        if self._receive_callback is not None:
            self._receive_callback(data)

    async def wait_until_disconnect(self):
        await self.disconnect_event.wait()

    def on_disconnect(self, client: BleakClient):
        logging.debug("Disconnect...")
        self.disconnect_event.set()

    async def send(self, data):
        if self._client is not None:
            await self._client.write_gatt_char(NORDIC_UART_TX_CHARACTERISTIC, data)

    async def send_forever(self, data: bytes, sleep_time: float):
        while True:
            await self.send(data)
            await asyncio.sleep(sleep_time)


async def send_text(
    adapter: str,
    text: str,
    device: str,
    loop: bool,
    sleep_time: float,
    timeout: float,
    callback=None,
):
    """Send text over Bluetooth LE.

    This is a facade that handle also connection/disconnection.
    """
    client = Client(adapter, device)
    if callback is not None:
        client.set_receive_callback(callback)
    try:
        await client.connect(timeout)
        await client.start()
        if loop:
            await asyncio.gather(
                client.wait_until_disconnect(),
                client.send_forever(bytes(text, "utf-8"), sleep_time),
            )
        else:
            await client.send(bytes(text, "utf-8"))
    finally:
        await client.disconnect()


async def repl_loop(
    adapter: str,
    device: str,
    timeout: float,
):
    """Run a REPL over BLE.

    This is a facade that handle also connection/disconnection.
    """
    client = Client(adapter, device)
    try:
        repl = REPL(client)
        await client.connect(timeout)
        await client.start()
        await asyncio.gather(client.wait_until_disconnect(), repl.shell())
    except asyncio.CancelledError:
        pass
    except AssertionError:
        raise
    finally:
        await client.disconnect()
